import logging

import umbi
from umbi.datatypes import (
    DataType,
    AtomicType,
    SizedType,
    BOOL1,
    UINT64,
)

from .tar import TarReader, TarWriter

logger = logging.getLogger(__name__)


class TarDecoder(TarReader):
    """An auxiliary class to simplify reading files of specific types from a tarball."""

    def read_vector(self, filename: str, sized_type: SizedType, required: bool = False) -> list | None:
        """
        Read a file as a vector of values.
        :param filename: name of the file to read
        :param sized_type: type of the values in the file
        :param required: if True, raise an error if the file is not found
        """
        data = self.read_file(filename, required)
        if data is None:
            return None
        return umbi.binary.bytes_to_vector(data, sized_type)

    @staticmethod
    def truncate_bitvector(vector: list[bool], num_entries: int) -> list[bool]:
        """Truncate a bitvector if its length exceeds num_entries."""
        if len(vector) > num_entries:
            if any(vector[num_entries:]):
                logger.warning(
                    f"bitvector {len(vector)} exceeds expected length of {num_entries}, truncating and discarding non-False entries"
                )
            vector = vector[:num_entries]
        return vector

    def read_bitvector(self, filename: str, num_entries: int | None) -> list[bool]:
        """Read a bitvector and truncate it to num_entries if necessary."""
        vector = self.read_vector(filename, BOOL1, required=True)
        assert isinstance(vector, list)
        if num_entries is not None:
            vector = self.truncate_bitvector(vector, num_entries)
        return vector

    def read_vector_with_csr(
        self, filename: str, value_type: DataType, required: bool, filename_csr: str
    ) -> list | None:
        """
        Read a file containing a vector of values while using an accompanying CSR.
        :param filename: name of the main file to read
        :param value_type: value type
        :param required: if True, raise an error if the main file is not found
        :param filename_csr: name of the accompanying CSR file
        """
        data = self.read_file(filename, required)
        if data is None:
            return None
        chunk_csr = self.read_vector(filename_csr, UINT64, required=True)
        assert chunk_csr is not None
        chunk_ranges = umbi.datatypes.csr_to_ranges(chunk_csr)
        return umbi.binary.bytes_with_csr_to_vector(data, value_type, chunk_ranges=chunk_ranges)

    def read_strings(self, filename: str, required: bool, filename_csr: str) -> list[str] | None:
        """
        Read a file containing a vector of strings, using an accompanying CSR file.
        :param filename: name of the main file to read
        :param required: if True, raise an error if the main file is not found
        :param filename_csr: name of the accompanying CSR file
        """
        return self.read_vector_with_csr(filename, umbi.datatypes.AtomicType.STRING, required, filename_csr)


class TarEncoder(TarWriter):
    def add_vector(self, filename: str, sized_type: SizedType, vector: list | None, required: bool = False):
        """
        Add a file containing a vector of values.
        :param filename: name of the file to add
        :param sized_type: type of the values
        :param vector: vector to add
        :param required: whether the file is required
        :raise ValueError: if required is True and vector is None
        """
        if vector is None:
            if required:
                raise ValueError(f"missing required data for {filename}")
            return
        data_out = umbi.binary.vector_to_bytes(vector, sized_type)
        self.add_file(filename, data_out)

    @staticmethod
    def pad_bitvector(vector: list[bool], num_bytes: int) -> list[bool]:
        """Pad a bitvector with False entries to make its length a multiple of num_bytes*8."""
        num_bits = num_bytes * 8
        items_to_add = (num_bits - (len(vector) % num_bits)) % num_bits
        if items_to_add > 0:
            # logger.debug(
            #     f"padding bitvector with {items_to_add} False entries to align to a {num_bytes}-byte boundary"
            # )
            vector = vector + [False] * items_to_add
        return vector

    def add_bitvector(self, filename: str, vector: list[bool], required: bool = False, pad_to_8_bytes: bool = True):
        """
        Add a bitvector file.
        :param filename: name of the file to add
        :param vector: bitvector data to write
        :param required: if True, raise an error if vector is None
        :param pad_to_8_bytes: if True, pad the bitvector to a multiple of 8 bytes
        """
        if vector is not None and pad_to_8_bytes:
            vector = self.pad_bitvector(vector, 8)
        self.add_vector(filename, BOOL1, vector, required)

    def add_vector_with_csr(
        self, filename: str, sized_type: SizedType, vector: list | None, filename_csr: str, required: bool = False
    ):
        """
        Add a file containing a vector of values, with an accompanying CSR file if needed.
        :param filename: name of the main file to add
        :param sized_type: value type
        :param vector: vector data to write
        :param filename_csr: name of the accompanying CSR file
        :param required: if True, raise an error if vector is None
        """
        if vector is None:
            if required:
                raise ValueError(f"missing required data for {filename}")
            return
        data_out, chunk_csr = umbi.binary.vector_to_bytes_with_csr(vector, sized_type)
        self.add_file(filename, data_out)
        if chunk_csr is not None:
            self.add_vector(filename_csr, UINT64, chunk_csr, required=True)

    def add_strings(self, filename: str, vector: list[str] | None, filename_csr: str, required: bool = False):
        """
        Add a file containing a vector of strings, using an accompanying CSR file.
        :param filename: name of the main file to add
        :param vector: list of strings to add
        :param filename_csr: name of the accompanying CSR file
        :param required: if True, raise an error if vector is None
        """
        self.add_vector_with_csr(filename, SizedType(AtomicType.STRING), vector, filename_csr, required)



from enum import Enum


class EntityClass(str, Enum):
    """Entity type of a transition system the values are usually associated with."""

    STATES = "states"
    CHOICES = "choices"
    BRANCHES = "branches"
    OBSERVATIONS = "observations"
    PLAYERS = "players"



from umbi.datatypes import NumericPrimitive, NumericPrimitiveType, SizedType
from .doubles import bytes_to_double, double_to_bytes, bits_to_double, double_to_bits
from .integers import bytes_to_integer, integer_to_bytes, bits_to_integer, integer_to_bits
from .rationals import bytes_to_rational, rational_to_bytes, bits_to_rational, rational_to_bits
from fractions import Fraction
from bitstring import BitArray


def bytes_to_numeric_primitive(data: bytes, value_type: NumericPrimitiveType, little_endian: bool = True):
    """Convert a binary string to a single value of the given numeric primitive type."""
    if value_type == NumericPrimitiveType.INT:
        return bytes_to_integer(data, signed=True, little_endian=little_endian)
    elif value_type == NumericPrimitiveType.UINT:
        return bytes_to_integer(data, signed=False, little_endian=little_endian)
    elif value_type == NumericPrimitiveType.DOUBLE:
        return bytes_to_double(data, little_endian=little_endian)
    else:  # value_type == NumericPrimitiveType.RATIONAL:
        return bytes_to_rational(data, little_endian=little_endian)


def numeric_primitive_to_bytes(value: NumericPrimitive, sized_type: SizedType, little_endian: bool = True) -> bytes:
    """Convert a numeric primitive value to a bytestring."""
    value_type = sized_type.type
    assert isinstance(value_type, NumericPrimitiveType), "expected a NumericPrimitiveType"
    if value_type == NumericPrimitiveType.INT:
        assert isinstance(value, int), "expected an integer value"
        return integer_to_bytes(value, sized_type.size_bytes, signed=True, little_endian=little_endian)
    if value_type == NumericPrimitiveType.UINT:
        assert isinstance(value, int), "expected an integer value"
        return integer_to_bytes(value, sized_type.size_bytes, signed=False, little_endian=little_endian)
    elif value_type == NumericPrimitiveType.DOUBLE:
        assert isinstance(value, float), "expected a float value"
        return double_to_bytes(value, little_endian=little_endian)
    else:  # value_type == NumericPrimitiveType.RATIONAL:
        assert isinstance(value, Fraction), "expected a rational value"
        return rational_to_bytes(value, sized_type.size_bytes, little_endian=little_endian)


def bits_to_numeric_primitive(bits: BitArray, value_type: NumericPrimitiveType) -> NumericPrimitive:
    """Convert a BitArray to a single value of the given numeric primitive type."""
    if value_type == NumericPrimitiveType.INT:
        return bits_to_integer(bits, signed=True)
    elif value_type == NumericPrimitiveType.UINT:
        return bits_to_integer(bits, signed=False)
    elif value_type == NumericPrimitiveType.DOUBLE:
        return bits_to_double(bits)
    else:  # value_type == NumericPrimitiveType.RATIONAL:
        return bits_to_rational(bits)


def numeric_primitive_to_bits(value: NumericPrimitive, sized_type: SizedType) -> BitArray:
    """Convert a numeric primitive value to a fixed-length bit representation."""
    value_type = sized_type.type
    if value_type == NumericPrimitiveType.INT:
        assert isinstance(value, int), "expected an integer value"
        return integer_to_bits(value, sized_type.size_bits, signed=True)
    elif value_type == NumericPrimitiveType.UINT:
        assert isinstance(value, int), "expected an integer value"
        return integer_to_bits(value, sized_type.size_bits, signed=False)
    elif value_type == NumericPrimitiveType.DOUBLE:
        assert isinstance(value, float), "expected a float value"
        return double_to_bits(value)
    else:  # value_type == NumericPrimitiveType.RATIONAL:
        assert isinstance(value, Fraction), "expected a rational value"
        return rational_to_bits(value, sized_type.size_bits)



"""
Variable valuation schemas and classes.
"""

from dataclasses import dataclass, field
from marshmallow import fields, post_load, validate
from marshmallow_oneofschema.one_of_schema import OneOfSchema

import umbi
from .type import SizedTypeSchema

from .json_schema import (
    JsonSchema,
    JsonSchemaResult,
    FieldUint,
)


class ValuationPaddingSchema(JsonSchema):
    """Schema for padding fields."""

    padding = FieldUint(data_key="padding", required=True, validate=validate.Range(min=1))

    @post_load
    def make_object(self, data: dict, **kwargs) -> umbi.datatypes.StructPadding:
        """Create a Padding object from the deserialized data."""
        obj = super().make_object(data, **kwargs)
        return umbi.datatypes.StructPadding(padding=obj.padding)

    def dump(self, obj, *args, **kwargs):
        assert isinstance(obj, umbi.datatypes.StructPadding)
        obj_dict = {
            "padding": obj.padding,
        }
        return obj_dict


class ValuationAttributeSchema(JsonSchema):
    """Schema for variable fields."""

    name = fields.String(data_key="name", required=True)
    type = fields.Nested(SizedTypeSchema, data_key="type", required=True)
    is_optional = fields.Boolean(data_key="is-optional", required=False, load_default=False)
    lower = fields.Integer(data_key="lower", required=False)
    upper = fields.Integer(data_key="upper", required=False)
    offset = fields.Integer(data_key="offset", required=False)

    @post_load
    def make_object(self, data: dict, **kwargs) -> umbi.datatypes.StructAttribute:
        """Validate and create a Variable object from the deserialized data."""
        obj = super().make_object(data, **kwargs)
        return umbi.datatypes.StructAttribute(
            name=obj.name,
            sized_type=obj.type,
            is_optional=obj.is_optional,
            lower=getattr(obj, "lower", None),
            upper=getattr(obj, "upper", None),
            offset=getattr(obj, "offset", None),
        )

    def dump(self, obj, *args, **kwargs):
        obj_dict = {
            "name": obj.name,
            "type": SizedTypeSchema().dump(obj.sized_type),
            "is-optional": obj.is_optional,
            "lower": obj.lower,
            "upper": obj.upper,
            "offset": obj.offset,
        }
        return obj_dict


class ValuationFieldSchema(OneOfSchema):
    """Schema for valuation fields (padding or variable)."""

    type_schemas = {
        "padding": ValuationPaddingSchema,
        "attribute": ValuationAttributeSchema,
    }

    # custom discriminator field since the default 'type' field is used in ValuationAttributeSchema
    type_field = "_discriminator"

    def get_obj_type(self, obj):
        """Determine which schema to use based on the object's attributes."""
        if hasattr(obj, "padding"):
            return "padding"
        elif hasattr(obj, "name") and hasattr(obj, "type"):
            return "attribute"
        else:
            raise ValueError("Object must be either a padding or attribute namespace")

    def load(self, json_data, *args, **kwargs):
        """Add discriminator field before loading."""
        assert isinstance(json_data, dict)
        json_data = dict(
            json_data,
            _discriminator="padding" if "padding" in json_data else "attribute",
        )
        return super().load(json_data, *args, **kwargs)

    def dump(self, obj, *args, **kwargs):
        """Remove discriminator field after dumping."""
        result = super().dump(obj, *args, **kwargs)
        assert isinstance(result, dict)
        result.pop("_discriminator", None)
        return result


class ValuationClassSchema(JsonSchema):
    """Schema for variable valuations."""

    variables = fields.List(fields.Nested(ValuationFieldSchema), data_key="variables", required=True)

    @post_load
    def make_object(self, data: dict, **kwargs) -> umbi.datatypes.StructType:
        obj = super().make_object(data, **kwargs)
        return umbi.datatypes.StructType(
            fields=obj.variables,
        )

    def dump(self, obj, *args, **kwargs):
        assert isinstance(obj, umbi.datatypes.StructType)
        variables = []
        for f in obj.fields:
            if isinstance(f, umbi.datatypes.StructPadding):
                variables.append(ValuationPaddingSchema().dump(f))
            else:
                assert isinstance(f, umbi.datatypes.StructAttribute)
                variables.append(ValuationAttributeSchema().dump(f))
        obj_dict = {
            "variables": variables,
        }
        return obj_dict


class ValuationDescriptionSchema(JsonSchema):
    unique = fields.Boolean(data_key="unique", required=True)
    num_strings = FieldUint(data_key="#strings", required=False, validate=validate.Range(min=1))
    classes = fields.List(
        fields.Nested(ValuationClassSchema), data_key="classes", required=True, validate=validate.Length(min=1)
    )

    @classmethod
    def schema_class(cls) -> type:
        return ValuationDescription


@dataclass
class ValuationDescription(JsonSchemaResult):
    unique: bool = False
    num_strings: int | None = None
    classes: list[umbi.datatypes.StructType] = field(default_factory=list)

    @classmethod
    def class_schema(cls) -> type:
        return ValuationDescriptionSchema


class ValuationsSchema(fields.Dict):
    """Schema for variable valuations as a dictionary."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            keys=fields.String(validate=validate.OneOf(["states", "choices", "branches", "observations", "players"])),
            values=fields.Nested(ValuationDescriptionSchema),
            *args,
            **kwargs,
        )



"""Utilities for (de)serializing intervals."""

from umbi.datatypes import SizedType, Interval, IntervalType, interval_base_type
from .utils import split_bytes_half

from .numeric_primitives import bytes_to_numeric_primitive, numeric_primitive_to_bytes


def bytes_to_interval(data: bytes, value_type: IntervalType, little_endian: bool = True) -> Interval:
    lower, upper = split_bytes_half(data)
    base_type = interval_base_type(value_type)
    lower = bytes_to_numeric_primitive(lower, base_type, little_endian)
    upper = bytes_to_numeric_primitive(upper, base_type, little_endian)
    return Interval(lower, upper)


def interval_to_bytes(interval: Interval, sized_type: SizedType, little_endian: bool = True) -> bytes:
    assert isinstance(sized_type.type, IntervalType), "expected an IntervalType"
    base_type = interval_base_type(sized_type.type)
    # TODO ensure both rationals use the same size for numerator and denominator
    assert sized_type.size_bits % 2 == 0, "expected an even-sized SizedType for interval"
    sized_base_type = SizedType(type=base_type, size_bits=sized_type.size_bits // 2)
    lower = numeric_primitive_to_bytes(interval.left, sized_base_type, little_endian)
    upper = numeric_primitive_to_bytes(interval.right, sized_base_type, little_endian)
    return lower + upper



"""
umbi.ats: Collection of interfaces over a umbfile.
"""

from .annotations import (
    Annotation,
    AtomicPropositionAnnotation,
    ObservationAnnotation,
    RewardAnnotation,
)
from .explicit_ats import ExplicitAts, TimeType
from .model_info import ModelInfo
from .variable_valuations import Variable, VariableValuations, EntityValuations, EntityClassValuations
from .entity_class import EntityClass

__all__ = [
    # annotation
    "Annotation",
    "AtomicPropositionAnnotation",
    "ObservationAnnotation",
    "RewardAnnotation",
    # model_info
    "ModelInfo",
    # explicit_ats
    "TimeType",
    "ExplicitAts",
    # variable_valuations
    "Variable",
    "VariableValuations",
    "EntityValuations",
    "EntityClassValuations",
    # entity_class
    "EntityClass",
]



"""
(De)serialization of common types.
"""

from umbi.datatypes import (
    ValueType,
    DataType,
    NumericPrimitive,
    SizedType,
    AtomicType,
    NumericPrimitiveType,
    Interval,
    IntervalType,
)

from .numeric_primitives import (
    bytes_to_numeric_primitive,
    numeric_primitive_to_bytes,
    bits_to_numeric_primitive,
    numeric_primitive_to_bits,
)
from .bitvectors import bits_to_bool, bool_to_bits
from .intervals import bytes_to_interval, interval_to_bytes
from .strings import bytes_to_string, string_to_bytes
from bitstring import BitArray


def bytes_to_value(data: bytes, value_type: DataType, little_endian: bool = True) -> ValueType:
    """Convert a binary string to a single value of the given common type."""
    if isinstance(value_type, AtomicType):
        if value_type == AtomicType.BOOL:
            return any(b != 0 for b in data)
        else:  # value_type == AtomicType.STRING:
            return bytes_to_string(data)
    else:  # umbi.datatypes.is_numeric_type(value_type):
        if isinstance(value_type, NumericPrimitiveType):
            return bytes_to_numeric_primitive(data, value_type, little_endian=little_endian)
        else:  # umbi.datatypes.is_interval_type(value_type):
            assert isinstance(value_type, IntervalType), "expected an interval type"
            return bytes_to_interval(data, value_type, little_endian=little_endian)


def value_to_bytes(value: ValueType, sized_type: SizedType, little_endian: bool = True) -> bytes:
    """Convert a value of a given type to a bytestring."""
    type = sized_type.type
    if isinstance(type, AtomicType):
        if type == AtomicType.BOOL:
            assert isinstance(value, bool), "expected a boolean value"
            return bytes([1 if value else 0])
        else:  # type == AtomicType.STRING:
            assert isinstance(value, str), "expected a string value"
            return string_to_bytes(value)
    else:  # umbi.datatypes.is_numeric_type(type):
        if isinstance(type, NumericPrimitiveType):
            assert isinstance(value, NumericPrimitive), "expected a numeric value"
            return numeric_primitive_to_bytes(value, sized_type, little_endian=little_endian)
        else:  # umbi.datatypes.is_interval_type(type):
            assert isinstance(type, IntervalType), "expected an interval type"
            assert isinstance(value, Interval), "expected an interval value"
            return interval_to_bytes(value, sized_type, little_endian=little_endian)


def bits_to_value(bits: BitArray, value_type: DataType) -> ValueType:
    """Convert a BitArray to a single value of the given common type."""
    if isinstance(value_type, AtomicType):
        if value_type == AtomicType.BOOL:
            return bits_to_bool(bits)
        else:  # value_type == AtomicType.STRING:
            raise ValueError("cannot unpack string from bits")
    else:  # umbi.datatypes.is_numeric_type(value_type):
        if isinstance(value_type, NumericPrimitiveType):
            return bits_to_numeric_primitive(bits, value_type)
        else:  # umbi.datatypes.is_interval_type(value_type):
            raise ValueError("cannot unpack interval from bits")


def value_to_bits(value: ValueType, sized_type: SizedType) -> BitArray:
    """Convert a value of a given type to a fixed-length bit representation."""
    type = sized_type.type
    if isinstance(type, AtomicType):
        if type == AtomicType.BOOL:
            assert isinstance(value, bool), "expected a boolean value"
            return bool_to_bits(value, sized_type.size_bits)
        else:  # type == AtomicType.STRING:
            raise ValueError("cannot pack string to bits")
    else:  # umbi.datatypes.is_numeric_type(type):
        if isinstance(type, NumericPrimitiveType):
            assert isinstance(value, NumericPrimitive), "expected a numeric value"
            return numeric_primitive_to_bits(value, sized_type)
        else:  # umbi.datatypes.is_interval_type(type):
            raise ValueError("cannot pack interval to bits")



"""
Interval datatypes.
"""

from fractions import Fraction
from .numeric_primitive import (
    NumericPrimitiveType,
    NumericPrimitive,
    common_numeric_primitive_type,
    promote_numeric_primitive_to,
)
import enum

""" Interval types. """


class IntervalType(str, enum.Enum):
    INT = "int-interval"
    UINT = "uint-interval"
    DOUBLE = "double-interval"
    RATIONAL = "rational-interval"


def interval_base_type(type: IntervalType) -> NumericPrimitiveType:
    return {
        IntervalType.INT: NumericPrimitiveType.INT,
        IntervalType.UINT: NumericPrimitiveType.UINT,
        IntervalType.DOUBLE: NumericPrimitiveType.DOUBLE,
        IntervalType.RATIONAL: NumericPrimitiveType.RATIONAL,
    }[type]


def base_to_interval_type(type: NumericPrimitiveType) -> IntervalType:
    return {
        NumericPrimitiveType.INT: IntervalType.INT,
        NumericPrimitiveType.UINT: IntervalType.UINT,
        NumericPrimitiveType.DOUBLE: IntervalType.DOUBLE,
        NumericPrimitiveType.RATIONAL: IntervalType.RATIONAL,
    }[type]


class Interval:
    """Represents a numeric interval where left <= right."""

    def __init__(self, left: NumericPrimitive, right: NumericPrimitive) -> None:
        self.left = left
        self.right = right
        self.validate()

    def validate(self) -> None:
        if not isinstance(self.left, NumericPrimitive):
            raise ValueError(f"expected numeric left bound, got: {self.left}")
        if not isinstance(self.right, NumericPrimitive):
            raise ValueError(f"expected numeric right bound, got: {self.right}")
        if not self.left <= self.right:
            raise ValueError(f"expected {self.left} <=  {self.right}")

    def __str__(self) -> str:
        return f"interval[{self.left},{self.right}]"

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, NumericPrimitive):
            other = Interval(other, other)
        if not isinstance(other, Interval):
            return False
        return self.left == other.left and self.right == other.right

    def __contains__(self, value: NumericPrimitive) -> bool:
        """Check if a numeric value is within the interval."""
        return self.left <= value <= self.right

    @property
    def base_type(self) -> NumericPrimitiveType:
        """Get the base numeric type of the interval."""
        if isinstance(self.left, Fraction) or isinstance(self.right, Fraction):
            return NumericPrimitiveType.RATIONAL
        elif isinstance(self.left, float) or isinstance(self.right, float):
            return NumericPrimitiveType.DOUBLE
        elif isinstance(self.left, int) or isinstance(self.right, int):
            return NumericPrimitiveType.INT
        else:
            raise ValueError(f"cannot determine base type of interval: {self}")

    @property
    def type(self) -> IntervalType:
        """Get the interval type of the interval."""
        return base_to_interval_type(self.base_type)


# def interval_type_of(value: object) -> IntervalType:
#     """Determine the interval type of a given value."""
#     if not isinstance(value, Interval):
#         raise ValueError(f"cannot match value to an interval type: {value}")
#     return value.type()


# def is_instance_of_interval_type(value: object, type: IntervalType) -> bool:
#     """Check if a value is an instance of the given interval type."""
#     return isinstance(value, Interval) and get_instance_interval_type(value) == type


def common_interval_type(types: set[IntervalType]) -> IntervalType:
    """Determine the common interval type from a set of interval types. Used for type promotion."""
    assert len(types) > 0, "cannot determine common numeric type of empty set"
    assert all(isinstance(t, IntervalType) for t in types), f"non-interval types found in set: {types}"
    base_types = {interval_base_type(t) for t in types}
    common_base_type = common_numeric_primitive_type(base_types)
    return base_to_interval_type(common_base_type)


def promote_interval_to(value: Interval, target_type: IntervalType) -> Interval:
    """Promote an interval to the target interval type."""
    target_base_type = interval_base_type(target_type)
    return Interval(
        promote_numeric_primitive_to(value.left, target_base_type),
        promote_numeric_primitive_to(value.right, target_base_type),
    )



"""
umbi.binary: Utilities for (de)serializing basic types.
"""

from .common import bytes_to_value, value_to_bytes
from .sequences import bytes_to_vector, bytes_with_csr_to_vector, vector_to_bytes, vector_to_bytes_with_csr

__all__ = [
    "value_to_bytes",
    "bytes_to_value",
    "bytes_to_vector",
    "bytes_with_csr_to_vector",
    "vector_to_bytes",
    "vector_to_bytes_with_csr",
]



"""
(De)serializers for integers.
"""

from bitstring import BitArray


def num_bits_for_integer(value: int, signed: bool = True, round_to_8: bool = True) -> int:
    """
    Return the number of bits needed to represent an integer value.
    :param signed: if True, the integer is signed
    :param round_to_8: if True, the number of bits is rounded up to the nearest multiple of 8
    """
    if not signed:
        num_bits = value.bit_length()
    else:
        if value >= 0:
            num_bits = value.bit_length()
        else:
            num_bits = (-value - 1).bit_length()
        num_bits += 1  # add sign bit
    if round_to_8 and num_bits % 8 != 0:
        num_bits += 8 - (num_bits % 8)
    return num_bits


def num_bytes_for_integer(value: int, signed: bool = True, round_to_8: bool = True) -> int:
    """
    Return the number of bytes needed to represent an integer value.
    :param round_to_8: if True, the number of bytes is rounded up to the nearest multiple of 8
    """
    num_bytes = num_bits_for_integer(value, signed=signed, round_to_8=round_to_8) // 8  # round up to full bytes
    if round_to_8 and num_bytes % 8 != 0:
        num_bytes += 8 - (num_bytes % 8)
    return num_bytes


def assert_integer_fits(
    value: int,
    signed: bool = True,
    num_bits: int | None = None,
    num_bytes: int | None = None,
):
    """Assert that an integer fits in the given number of bits/bytes."""
    assert (num_bits is None) != (num_bytes is None), "either num_bits or num_bytes must be provided, but not both"
    if num_bits is None:
        assert num_bytes is not None  # for mypy
        num_bits = num_bytes * 8
    if signed:
        max_value = (1 << (num_bits - 1)) - 1
        min_value = -(1 << (num_bits - 1))
    else:
        max_value = (1 << num_bits) - 1
        min_value = 0
    if not (min_value <= value <= max_value):
        raise ValueError(
            f"integer value {value} is out of range for a {num_bits}-bit {'signed' if signed else 'unsigned'} integer [{min_value}, {max_value}]"
        )


def bytes_to_integer(data: bytes, signed: bool = True, little_endian: bool = True) -> int:
    return int.from_bytes(data, byteorder="little" if little_endian else "big", signed=signed)


def integer_to_bytes(value: int, num_bytes: int, signed: bool = True, little_endian: bool = True) -> bytes:
    assert_integer_fits(value, signed=signed, num_bytes=num_bytes)
    return value.to_bytes(num_bytes, byteorder="little" if little_endian else "big", signed=signed)


def bits_to_integer(bits: BitArray, signed: bool = True) -> int:
    """Convert a BitArray to a single integer value."""
    if signed:
        return bits.int
    else:
        return bits.uint


def integer_to_bits(value: int, num_bits: int, signed: bool = True) -> BitArray:
    """Convert a single integer value to a fixed-length bit representation."""
    assert_integer_fits(value, signed=signed, num_bits=num_bits)
    if signed:
        return BitArray(int=value, length=num_bits)
    else:
        return BitArray(uint=value, length=num_bits)



"""
Auxiliary vector operations.
"""

from dataclasses import dataclass

from .datatype import DataType, ValueType, common_datatype, datatype_of, promote_value_to
from .sized_type import SizedType, UINT64
from .atomic import AtomicType

from collections.abc import Iterable

# TODO CSR vector as a dedicated class
# TODO add vector of ranges (for CSR?)


@dataclass
class VectorType:
    type: SizedType


"""Alias for a CSR vector type."""
CSR_VECTOR = VectorType(UINT64)

"""Alias for a vector of booleans."""
BIT_VECTOR = VectorType(SizedType(AtomicType.BOOL))


def assert_is_list(vector: object):
    if not isinstance(vector, list):
        raise TypeError(f"expected a list/vector, got {type(vector)}")


def is_vector_ranges(ranges: list[tuple[int, int]]) -> bool:
    """Check if a vector is a list of ranges."""
    if len(ranges) == 0:
        return False

    def is_interval(x: tuple[int, int]) -> bool:
        return isinstance(x, tuple) and len(x) == 2 and all(isinstance(y, int) for y in x) and x[0] <= x[1]

    if not all(is_interval(interval) for interval in ranges):
        return False
    for i in range(len(ranges) - 1):
        if ranges[i][1] != ranges[i + 1][0]:
            return False
    return True


def is_vector_csr(vector: list[int]) -> bool:
    """Check if a vector is a CSR row start index vector."""
    if len(vector) < 2:
        return False
    if not all(isinstance(x, int) for x in vector):
        return False
    if vector[0] != 0:
        return False
    if not all(vector[i] <= vector[i + 1] for i in range(len(vector) - 1)):
        return False
    return True


def csr_to_ranges(csr: list[int]) -> list[tuple[int, int]]:
    """Convert row start indices to ranges."""
    assert is_vector_csr(csr), "input is not a valid CSR vector"
    ranges = [(csr[i], csr[i + 1]) for i in range(len(csr) - 1)]
    return ranges


def ranges_to_csr(ranges: list[tuple[int, int]]) -> list[int]:
    """Convert ranges to CSR row start indices."""
    assert is_vector_ranges(ranges), "input is not a valid ranges vector"
    csr = [interval[0] for interval in ranges]
    csr.append(ranges[-1][-1])
    return csr


def collection_element_types(collection: Iterable[ValueType]) -> set[DataType]:
    """Determine the set of common types of elements in the collection."""
    return set([datatype_of(x) for x in collection])


def collection_element_type(collection: Iterable[ValueType]) -> DataType:
    """Determine the common type of elements in the collection. Raises an error if multiple types are found."""
    types = collection_element_types(collection)
    if len(types) != 1:
        raise ValueError(f"collection has multiple element types: {types}")
    return types.pop()


def common_collection_element_type(collection: Iterable[ValueType]) -> DataType:
    """Determine the common type to which all elements in the collection can be promoted."""
    return common_datatype(collection_element_types(collection))


def promote_vector_to(vector: list[ValueType], target_type: DataType) -> list[ValueType]:
    """Promote a vector of values to the target type."""
    return [promote_value_to(v, target_type) for v in vector]



from dataclasses import dataclass
from .atomic import AtomicType
from .numeric_primitive import NumericPrimitiveType
from .interval import IntervalType
from .datatype import DataType


def atomic_type_default_size(type: AtomicType) -> int:
    return {
        AtomicType.BOOL: 1,
        AtomicType.STRING: 64,
    }[type]


def numeric_primitive_type_default_size(type: NumericPrimitiveType) -> int:
    return {
        NumericPrimitiveType.INT: 64,
        NumericPrimitiveType.UINT: 64,
        NumericPrimitiveType.DOUBLE: 64,
        NumericPrimitiveType.RATIONAL: 128,
    }[type]


def interval_type_default_size(type: IntervalType) -> int:
    return {
        IntervalType.INT: 128,
        IntervalType.UINT: 128,
        IntervalType.DOUBLE: 128,
        IntervalType.RATIONAL: 256,
    }[type]


def type_default_size(type: DataType) -> int:
    if isinstance(type, AtomicType):
        return atomic_type_default_size(type)
    elif isinstance(type, NumericPrimitiveType):
        return numeric_primitive_type_default_size(type)
    else:  # isinstance(type, IntervalType)
        return interval_type_default_size(type)


def validate_atomic_type_size(type: AtomicType, size: int) -> None:
    if type == AtomicType.STRING:
        if size != 64:
            raise ValueError(f"{type.value} size must be 64")


def validate_numeric_primitive_type_size(type: NumericPrimitiveType, size: int) -> None:
    if type == NumericPrimitiveType.DOUBLE:
        if size != 64:
            raise ValueError(f"{type.value} size must be 64")
    if type == NumericPrimitiveType.RATIONAL:
        if size % 2 != 0:
            raise ValueError(f"{type.value} size must be a multiple of 2")


def validate_interval_type_size(type: IntervalType, size: int) -> None:
    if type in {IntervalType.INT, IntervalType.UINT, IntervalType.DOUBLE, IntervalType.RATIONAL}:
        if size % 2 != 0:
            raise ValueError(f"{type.value} size must be a multiple of 2")
        if type == IntervalType.RATIONAL and size % 4 != 0:
            raise ValueError(f"{type.value} size must be a multiple of 4")


def validate_type_size(type: DataType, size: int) -> None:
    if isinstance(type, AtomicType):
        validate_atomic_type_size(type, size)
    elif isinstance(type, NumericPrimitiveType):
        validate_numeric_primitive_type_size(type, size)
    else:  # isinstance(type, IntervalType)
        validate_interval_type_size(type, size)


@dataclass
class SizedType:
    """
    Represents a type definition for common types.
    """

    type: DataType
    size_bits: int  # size in bits

    def __init__(self, type: DataType, size_bits: int | None = None):
        self.type = type
        if size_bits is None:
            size_bits = type_default_size(type)
        self.size_bits = size_bits

    def __post_init__(self):
        self.validate()

    def validate(self):
        assert self.size_bits > 0, "size must be positive"
        validate_type_size(self.type, self.size_bits)

    @property
    def is_byte_aligned(self) -> bool:
        return self.size_bits % 8 == 0

    @property
    def size_bytes(self) -> int:
        return (self.size_bits + 7) // 8  # ceiling division


# Common sized types used in umbi
BOOL1 = SizedType(AtomicType.BOOL, 1)
UINT16 = SizedType(NumericPrimitiveType.UINT, 16)
UINT32 = SizedType(NumericPrimitiveType.UINT, 32)
UINT64 = SizedType(NumericPrimitiveType.UINT, 64)



"""
Atomic datatypes. "Atomic" is an umbrella term for primitive types that are neither numeric nor interval types.
#TODO: expand to JSONs?
"""

import enum

""" Atomic types. """


class AtomicType(str, enum.Enum):
    BOOL = "bool"
    STRING = "string"


""" Alias for atomic objects. """
Atomic = bool | str


def atomic_type_of(value: Atomic) -> AtomicType:
    """Determine the primitive type of a given value."""
    if isinstance(value, bool):
        return AtomicType.BOOL
    else:  # isinstance(value, str):
        return AtomicType.STRING


# def is_of_atomic_type(value: object, type: AtomicType) -> bool:
#     """Check if a value is an instance of the given atomic type."""
#     return atomic_type_of(value) == type



"""
Utilities for packing and unpacking composite datatypes (structs).
"""

from bitstring import BitArray

from umbi.datatypes import StructType, StructAttribute, StructPadding, AtomicType, ValueType, UINT64
from .common import bits_to_value, value_to_bits
from .utils import split_bytes


class StructUnpacker:
    """Utility class for unpacking composite datatypes from a bytestring."""

    def __init__(self, bytestring: bytes):
        self.bytestring = bytestring  # input bytestring, little-endian order
        self.buffer = BitArray()  # bit buffer, MSB at [0]

    def assert_buffer_empty(self):
        if len(self.buffer) > 0:
            raise RuntimeError("expected the buffer to be empty")

    def align_buffer(self, num_bits: int):
        """Ensure the buffer has the required number ofbits."""
        if len(self.buffer) >= num_bits:
            return
        num_bytes_needed = (num_bits - len(self.buffer) + 7) // 8
        assert len(self.bytestring) >= num_bytes_needed, "not enough data to fill the buffer"
        for _ in range(num_bytes_needed):  # read one byte at a time since the bytestring is little-endian
            next_bytes, self.bytestring = split_bytes(self.bytestring, 1)
            self.buffer = BitArray(bytes=next_bytes) + self.buffer  # prepend to the buffer to keep MSB at [0]

    def extract_from_buffer(self, num_bits: int):
        """Extract the given number of bits from the buffer."""
        assert num_bits >= 0
        self.align_buffer(num_bits)
        bits = self.buffer[-num_bits:]  # remove from the end (LSB side)
        self.buffer = self.buffer[:-num_bits]
        return bits

    def skip_padding(self, field: StructPadding):
        self.extract_from_buffer(field.padding)

    def unpack_attribute(self, field: StructAttribute) -> ValueType | None:
        """
        Unpack a single field from the buffer.
        :return: the unpacked value, or None if the field is optional and not present
        """
        sized_type = field.sized_type
        if field.sized_type.type == AtomicType.STRING:
            # a string is stored as an index to the list of strings
            sized_type = UINT64
        bits = self.extract_from_buffer(sized_type.size_bits)
        if field.is_optional:
            is_present_bit = bits[-1]  # remove from the end (LSB side)
            bits = bits[:-1]
            if not is_present_bit:
                return None
        return bits_to_value(bits, sized_type.type)

    def unpack_struct(self, value_type: StructType) -> dict[str, ValueType | None]:
        name_value = dict()
        for field in value_type:
            if isinstance(field, StructPadding):
                self.skip_padding(field)
                continue
            name_value[field.name] = self.unpack_attribute(field)
        self.assert_buffer_empty()
        return name_value


class StructPacker:
    """Utility class for packing structs into a bytestring."""

    def __init__(self):
        self.buffer = BitArray()  # bit buffer, MSB at [0]
        self.bytestring = bytes()  # output bytestring, little-endian order

    def assert_buffer_empty(self):
        if len(self.buffer) > 0:
            raise RuntimeError("expected the buffer to be empty")

    def flush_buffer(self):
        """Flush full bytes from the buffer to the bytestring."""
        while len(self.buffer) >= 8:  # append one byte at a time so that the bytestring is little-endian
            self.buffer, bits = (
                self.buffer[:-8],
                self.buffer[-8:],
            )  # get new bits from the end (LSB side)
            new_bytes = bits.tobytes()
            self.bytestring += new_bytes

    def append_to_buffer(self, bits: BitArray):
        """Append bits to the buffer flush full bytes to the bytestring."""
        self.buffer = bits + self.buffer  # prepend to the start (MSB side)
        self.flush_buffer()

    def add_padding(self, field: StructPadding):
        """Add padding bits to the buffer."""
        self.append_to_buffer(BitArray(uint=0, length=field.padding))

    def pack_attribute(self, field: StructAttribute, value: ValueType | None):
        """Pack a single attribute into the buffer or the bytestring."""
        sized_type = field.sized_type
        if field.sized_type.type == AtomicType.STRING:
            # a string is stored as an index to the list of strings
            sized_type = UINT64
        bits = None
        if value is None:
            assert field.is_optional, "only optional fields can have None value"
            # add an arbitrary value with the presence bit set to 0
            bits = value_to_bits(0, sized_type)
            bits = bits + BitArray(uint=0, length=1)  # set presence bit to 0
        else:
            bits = value_to_bits(value, sized_type)
            if field.is_optional:
                presence_flag = BitArray(uint=1, length=1)  # set presence bit to 1
                bits = bits + presence_flag
        if field.is_optional:
            is_present_bit = bits[-1]  # remove from the end (LSB side)
            bits = bits[:-1]
            if not is_present_bit:
                return None
        self.append_to_buffer(bits)

    def pack_struct(self, value_sized_type: StructType, values: dict[str, ValueType | None]) -> bytes:
        assert value_sized_type.is_byte_aligned, "cannot pack a struct that is not byte-aligned"
        for field in value_sized_type:
            if isinstance(field, StructPadding):
                self.add_padding(field)
            else:
                assert isinstance(field, StructAttribute)
                self.pack_attribute(field, values[field.name])
        self.assert_buffer_empty()
        return self.bytestring


def struct_unpack(bytestring: bytes, value_type: StructType) -> dict[str, ValueType | None]:
    """Unpack a BitArray to a composite datatype."""
    return StructUnpacker(bytestring).unpack_struct(value_type)


def struct_pack(values: dict[str, ValueType | None], value_type: StructType) -> bytes:
    """Convert a composite datatype to a BitArray."""
    return StructPacker().pack_struct(value_type, values)



"""
Utilities for (de)serializing fractions.
"""

from fractions import Fraction

from .integers import (
    num_bytes_for_integer,
    integer_to_bytes,
    bytes_to_integer,
)
from .utils import split_bytes_half
from bitstring import BitArray

# Convention: a normalized rational has a non-negative denominator.
# Rationals are represented as two integers of equal lengths: a signed numerator and an unsigned denominator.


def normalize_rational(value: Fraction) -> Fraction:
    """Ensure that the denominator of a fraction is non-negative."""
    if value.denominator < 0:
        value = Fraction(-value.numerator, -value.denominator)
    return value


# def num_bits_for_rational(value: Fraction) -> int:
#     """Calculate the number of bits needed to represent a rational number."""
#     numerator_size = num_bits_for_integer(value.numerator, signed=True, round_to_8=False)
#     denominator_size = num_bits_for_integer(value.denominator, signed=False, round_to_8=False)
#     total_size = max(numerator_size, denominator_size) * 2
#     return total_size


def num_bytes_for_rational(value: Fraction) -> int:
    """Calculate the number of bytes needed to represent a rational number."""
    # rounding up to a number of bytes of multiple of 8
    numerator_size = num_bytes_for_integer(value.numerator, signed=True, round_to_8=True)
    denominator_size = num_bytes_for_integer(value.denominator, signed=False, round_to_8=True)
    return max(numerator_size, denominator_size) * 2


def bytes_to_rational(data: bytes, little_endian: bool = True) -> Fraction:
    """Convert a bytestring to a fraction. The bytestring must have even length, with the first half representing the numerator as a signed integer and the second half representing the denominator as an unsigned integer."""
    numerator_bytes, denominator_bytes = split_bytes_half(data)
    numerator = bytes_to_integer(numerator_bytes, signed=True, little_endian=little_endian)
    denominator = bytes_to_integer(denominator_bytes, signed=False, little_endian=little_endian)
    return Fraction(numerator, denominator)


def rational_to_bytes(value: Fraction, size_bytes: int, little_endian: bool = True) -> bytes:
    """
    Convert a fraction to a bytestring. Both numberator and denominator are encoded as signed and unsigned integers, respectively, and have the same size.
    :param size_bytes: maximum size in bytes for numerator/denominator
    """
    value = normalize_rational(value)
    assert size_bytes % 2 == 0, "size_bytes must be even to encode both numerator and denominator"
    term_size_bytes = size_bytes // 2
    minimal_term_size = num_bytes_for_rational(value) // 2
    if term_size_bytes < minimal_term_size:
        raise ValueError(
            f"term_size {term_size_bytes} is too small to represent the rational {value}, which requires at least {minimal_term_size} bytes per term"
        )
    numerator_bytes = integer_to_bytes(
        value.numerator, num_bytes=term_size_bytes, signed=True, little_endian=little_endian
    )
    denominator_bytes = integer_to_bytes(
        value.denominator,
        num_bytes=term_size_bytes,
        signed=False,
        little_endian=little_endian,
    )
    return numerator_bytes + denominator_bytes


def bits_to_rational(bits: BitArray) -> Fraction:
    assert len(bits) % 8 == 0, "rational bit length must be a multiple of 8"
    # rational are always byte-alignd
    bytestring = bits.tobytes()
    return bytes_to_rational(bytestring)


def rational_to_bits(value: Fraction, size_bytes: int) -> BitArray:
    bytestring = rational_to_bytes(value, size_bytes=size_bytes)
    return BitArray(bytestring)



import logging
from dataclasses import dataclass, field, fields
from enum import Enum
from typing import Iterable

from .entity_class import EntityClass
from .annotations import (
    Annotation,
    RewardAnnotation,
    AtomicPropositionAnnotation,
    ObservationAnnotation,
)
from .model_info import ModelInfo
from .variable_valuations import EntityClassValuations

logger = logging.getLogger(__name__)


class TimeType(str, Enum):
    """Time types for transition systems."""

    DISCRETE = "discrete"
    STOCHASTIC = "stochastic"
    URGENT_STOCHASTIC = "urgent-stochastic"


@dataclass
class ExplicitAts:
    """
    Explicit container for an annotated transition system (ATS). All the attributes can be freely manipulated.
    Method validate() can be used to check for consistency.
    """

    model_info: ModelInfo | None = None
    """Information about the model."""
    time: TimeType = TimeType.DISCRETE
    """Type of time."""
    num_states: int = 1
    """Number of states. Must be > 0."""

    num_players: int = 0
    """Number of players."""
    state_to_player: list[int] | None = None
    """State-to-player mapping. Must be set if num_players >= 1."""
    player_to_name: list[str] | None = None
    """Player-to-name mapping. Can be set only if num_players >= 1."""

    num_initial_states: int = 0
    """Number of initial states. Must be consistent with state_is_initial."""
    state_is_initial: list[bool] = field(default_factory=list)
    """State-to-whether-initial mapping."""

    num_choices: int = 0
    """Number of choices."""
    state_to_choice: list[int] | None = None
    """CSR list of length num_states + 1. Must be set if num_choices > 0."""

    num_branches: int = 0
    """Number of branches."""
    choice_to_branches: list[int] | None = None
    """CSR list of length num_choices + 1. Must be set if num_branches > 0."""

    branch_to_target: list[int] | None = None
    """Branch-to-target-state mapping. Must be set if num_branches > 0."""
    branch_to_probability: list | None = None
    """Branch-to-probability mapping. Must be set if num_branches > 0. Can contain arbitrary Numeric values."""

    state_is_markovian: list[bool] | None = None
    """State-to-whether-markovian mapping. Must be set if time is TimeType.STOCHASTIC."""
    state_to_exit_rate: list | None = None
    """State-to-exit-rate mapping. Must be set if time is TimeType.STOCHASTIC. Can contain arbitrary Numeric values."""

    num_choice_actions: int = 0
    """Number of actions associated with choices."""
    choice_to_choice_action: list[int] | None = None
    """Choice-to-choice-action mapping. Must be set if num_choice_actions > 0."""
    choice_action_to_name: list[str] | None = None
    """Choice-action-to-string mapping. Can only be set if num_choice_actions > 0."""

    num_branch_actions: int = 0
    """Number of actions associated with branches."""
    branch_to_branch_action: list[int] | None = None
    """Branch-to-branch-action mapping. Must be set if num_branch_actions > 0."""
    branch_action_to_name: list[str] | None = None
    """Branch-action-to-string mapping. Can only be set if num_branch_actions > 0."""

    annotations: dict[str, dict[str, Annotation]] = field(default_factory=dict)
    """
    Annotation category -> (annotation name -> annotation) mapping. Categories 'rewards' and 'aps' can be used, but
    must be of the type RewardAnnotation and AtomicPropositionAnnotation, respectively.
    """

    observation_annotation: ObservationAnnotation | None = None
    """Observation annotation."""

    variable_valuations: EntityClassValuations | None = None
    """EntityClassValuations associated with the ATS."""

    def __str__(self) -> str:
        lines = [f"{self.__class__.__name__}("]
        for dataclass_field in fields(self):
            lines.append(f"  {dataclass_field.name}={getattr(self, dataclass_field.name)!r}")
        lines.append(")")
        return "\n".join(lines)

    def __repr__(self) -> str:
        return self.__str__()

    # helper properties and methods

    @property
    def initial_states(self) -> list[int]:
        """Get the list of the initial states."""
        return [i for i, is_initial in enumerate(self.state_is_initial) if is_initial]

    def set_initial_states(self, initial_states: Iterable[int]):
        """Set the initial states."""
        self.state_is_initial = [False] * self.num_states
        for s in initial_states:
            if s >= self.num_states:
                raise ValueError(f"Invalid state {s}, must be < {self.num_states}.")
            self.state_is_initial[s] = True
        self.num_initial_states = len([is_initial for is_initial in self.state_is_initial if is_initial])

    def get_player_of_state(self, state: int) -> int:
        """Get the player controlling the given state."""
        if self.state_to_player is None:
            raise ValueError("state_to_player is not set")
        return self.state_to_player[state]

    @property
    def markovian_states(self) -> list[int]:
        """Get the list of the markovian states."""
        if self.state_is_markovian is None:
            raise ValueError("state_is_markovian is not set")
        return [i for i, is_markovian in enumerate(self.state_is_markovian) if is_markovian]

    def set_markovian_states(self, markovian_states: Iterable[int]):
        """Set the markovian states."""
        self.state_is_markovian = [False] * self.num_states
        for s in markovian_states:
            if s >= self.num_states:
                raise ValueError(f"Invalid state {s}, must be < {self.num_states}.")
            self.state_is_markovian[s] = True

    def state_choice_range(self, state: int) -> Iterable[int]:
        """Return the choice range of the given state."""
        if self.state_to_choice is None:
            raise ValueError("state_to_choice is not set")
        assert state <= self.num_states
        return range(self.state_to_choice[state], self.state_to_choice[state + 1])

    def choice_branch_range(self, choice: int) -> Iterable[int]:
        """Return the branch range of the given choice."""
        if self.choice_to_branches is None:
            raise ValueError("choice_to_branches is not set")
        assert choice <= self.num_choices
        return range(self.choice_to_branches[choice], self.choice_to_branches[choice + 1])

    ### Rewards. ###

    @property
    def reward_annotations(self) -> dict[str, RewardAnnotation]:
        """Get the reward annotations."""
        if "rewards" not in self.annotations:
            self.annotations["rewards"] = {}
        return self.annotations["rewards"]  # type: ignore

    @property
    def has_reward_annotations(self) -> bool:
        """Check if there are any reward annotations."""
        return len(self.reward_annotations) > 0

    @property
    def reward_annotation_names(self) -> list[str]:
        """Get the names of all reward annotations."""
        return list(self.reward_annotations.keys())

    def has_reward_annotation(self, name: str) -> bool:
        """Check if a reward annotation with the given name exists."""
        return name in self.reward_annotations

    def add_reward_annotation(self, annotation: RewardAnnotation):
        """Add a reward annotation."""
        if self.has_reward_annotation(annotation.name):
            raise ValueError(f"reward annotation with name {annotation.name} already exists")
        self.reward_annotations[annotation.name] = annotation

    def get_reward_annotation(self, name: str) -> RewardAnnotation:
        """Get the reward annotation with the given name."""
        if not self.has_reward_annotation(name):
            raise ValueError(f"reward annotation with name {name} does not exist")
        return self.reward_annotations[name]

    ### Atomic propositions. ###

    @property
    def ap_annotations(self) -> dict[str, AtomicPropositionAnnotation]:
        """Get the atomic proposition annotations."""
        if "aps" not in self.annotations:
            self.annotations["aps"] = {}
        return self.annotations["aps"]  # type: ignore

    @property
    def has_ap_annotations(self) -> bool:
        """Check if there are any atomic proposition annotations."""
        return len(self.ap_annotations) > 0

    @property
    def ap_annotation_names(self) -> list[str]:
        """Get the names of all atomic proposition annotations."""
        return list(self.ap_annotations.keys())

    def has_ap_annotation(self, name: str) -> bool:
        """Check if an atomic proposition annotation with the given name exists."""
        return name in self.ap_annotations

    def add_ap_annotation(self, annotation: AtomicPropositionAnnotation):
        """Add an atomic proposition annotation."""
        if self.has_ap_annotation(annotation.name):
            raise ValueError(f"atomic proposition annotation with name {annotation.name} already exists")
        self.ap_annotations[annotation.name] = annotation

    def get_ap_annotation(self, name: str) -> AtomicPropositionAnnotation:
        """Get the atomic proposition annotation with the given name."""
        if not self.has_ap_annotation(name):
            raise ValueError(f"atomic proposition annotation with name {name} does not exist")
        return self.ap_annotations[name]

    ### Observations. ###

    @property
    def num_observations(self) -> int:
        if self.observation_annotation is None:
            return 0
        return self.observation_annotation.num_observations

    @property
    def has_observations(self) -> bool:
        return self.num_observations > 0

    ### Variable valuations. ###

    @property
    def has_variable_valuations(self) -> bool:
        return self.variable_valuations is not None

    def validate(self):
        # TODO more checks

        if not self.num_states > 0:
            raise ValueError("expected num_states > 0")

        if self.num_players > 1:
            if self.state_to_player is None:
                raise ValueError("num_players > 1 but state_to_player is None")
            if len(self.state_to_player) != self.num_states:
                raise ValueError("expected len(state_to_player) == num_states")

        if len(self.state_is_initial) != self.num_states:
            raise ValueError("expected len(state_is_initial) == num_states")
        if self.num_initial_states != len(self.initial_states):
            raise ValueError("expected num_initial_states == len(initial_states)")

        if self.state_to_choice is not None:
            if len(self.state_to_choice) != self.num_states + 1:
                raise ValueError("expected len(state_to_choice) == num_states+1")
        if self.choice_to_branches is not None:
            if len(self.choice_to_branches) != self.num_choices + 1:
                raise ValueError("expected len(choice_to_branches) == num_choices+1")
        if self.branch_to_target is not None:
            if len(self.branch_to_target) != self.num_branches:
                raise ValueError("expected len(branch_to_target) == num_branches")

        for reward_annotation in self.reward_annotations.values():
            if not isinstance(reward_annotation, RewardAnnotation):
                raise ValueError(f"expected RewardAnnotation, got {type(reward_annotation)}")
            reward_annotation.validate()
        for ap_annotation in self.ap_annotations.values():
            if not isinstance(ap_annotation, AtomicPropositionAnnotation):
                raise ValueError(f"expected AtomicPropositionAnnotation, got {type(ap_annotation)}")
            ap_annotation.validate()
        if self.observation_annotation is not None:
            self.observation_annotation.validate()
        if self.has_variable_valuations:
            assert self.variable_valuations is not None
            for entity_class in self.variable_valuations.entity_classes:
                valuations = self.variable_valuations.get_valuations_for(entity_class)
                num_entities = {
                    EntityClass.STATES: self.num_states,
                    EntityClass.CHOICES: self.num_choices,
                    EntityClass.BRANCHES: self.num_branches,
                    EntityClass.OBSERVATIONS: self.num_observations,
                    EntityClass.PLAYERS: self.num_players,
                }[entity_class]
                if not valuations.num_entities == num_entities:
                    raise ValueError(
                        f"{entity_class} valuations has {valuations.num_entities} entities, expected {num_entities}"
                    )
                valuations.validate()

    def equal(self, other: object, debug=False) -> bool:
        if not isinstance(other, ExplicitAts):
            if debug:
                logger.debug("ExplicitAts.__eq__: other is not an ExplicitAts")
            return False
        equal = True
        for field_name in self.__dataclass_fields__:
            if getattr(self, field_name) != getattr(other, field_name):
                equal = False
                if not debug:
                    break
                logger.debug(f"ExplicitAts.__eq__: field {field_name} differs")
                logger.debug(f"  self: {getattr(self, field_name)}")
                logger.debug(f"  other: {getattr(other, field_name)}")
        return equal

    def __eq__(self, other):
        return self.equal(other)



"""
umbi.datatypes: Core datatype definitions and utilities.

This package contains the fundamental datatype classes and definitions used throughout umbi,
including intervals, composite types, vector utilities, and JSON operations.
"""

# atomic.py
from .atomic import AtomicType, Atomic, atomic_type_of

# datatype.py
from .datatype import DataType, ValueType, datatype_of, common_datatype, promote_value_to

# interval.py
from .interval import IntervalType, Interval, interval_base_type, base_to_interval_type

# json.py
from .json import (
    JsonPrimitive,
    JsonList,
    JsonDict,
    JsonLike,
    is_json_instance,
    json_remove_none_dict_values,
    json_to_string,
    string_to_json,
)

# struct.py
from .struct import StructPadding, StructAttribute, StructType

# vector.py
from .vector import (
    VectorType,
    assert_is_list,
    is_vector_ranges,
    is_vector_csr,
    csr_to_ranges,
    ranges_to_csr,
    collection_element_types,
    collection_element_type,
    common_collection_element_type,
    promote_vector_to,
)

# sized_type.py
from .sized_type import (
    SizedType,
    atomic_type_default_size,
    numeric_primitive_type_default_size,
    interval_type_default_size,
    type_default_size,
    validate_atomic_type_size,
    validate_numeric_primitive_type_size,
    validate_interval_type_size,
    validate_type_size,
    BOOL1,
    UINT16,
    UINT32,
    UINT64,
)

# numeric_primitive.py
from .numeric_primitive import (
    NumericPrimitiveType,
    NumericPrimitive,
    numeric_primitive_type_of,
    common_numeric_primitive_type,
    promote_numeric_primitive_to,
    is_integer_type,
    integer_type_signed,
)

# numeric.py
from .numeric import (
    NumericType,
    Numeric,
    is_numeric_primitive_type,
    is_interval_type,
    numeric_type_of,
    common_numeric_type,
    promote_numeric_to,
)

__all__ = [
    # atomic.py
    "AtomicType",
    "Atomic",
    "atomic_type_of",
    # datatype.py
    "DataType",
    "ValueType",
    "datatype_of",
    "common_datatype",
    "promote_value_to",
    "promote_vector_to",
    # interval.py
    "IntervalType",
    "Interval",
    "interval_base_type",
    "base_to_interval_type",
    # json.py
    "JsonPrimitive",
    "JsonList",
    "JsonDict",
    "JsonLike",
    "is_json_instance",
    "json_remove_none_dict_values",
    "json_to_string",
    "string_to_json",
    # struct.py
    "StructPadding",
    "StructAttribute",
    "StructType",
    # vector.py
    "VectorType",
    "assert_is_list",
    "is_vector_ranges",
    "is_vector_csr",
    "csr_to_ranges",
    "ranges_to_csr",
    "collection_element_types",
    "collection_element_type",
    "common_collection_element_type",
    # sized_type.py
    "SizedType",
    "atomic_type_default_size",
    "numeric_primitive_type_default_size",
    "interval_type_default_size",
    "type_default_size",
    "validate_atomic_type_size",
    "validate_numeric_primitive_type_size",
    "validate_interval_type_size",
    "validate_type_size",
    "BOOL1",
    "UINT16",
    "UINT32",
    "UINT64",
    # numeric_primitive.py
    "NumericPrimitiveType",
    "NumericPrimitive",
    "numeric_primitive_type_of",
    "common_numeric_primitive_type",
    "promote_numeric_primitive_to",
    "is_integer_type",
    "integer_type_signed",
    # numeric.py
    "NumericType",
    "Numeric",
    "is_numeric_primitive_type",
    "is_interval_type",
    "numeric_type_of",
    "common_numeric_type",
    "promote_numeric_to",
]



import logging
from typing import Iterable
from umbi.datatypes import (
    ValueType,
    DataType,
    collection_element_types,
    common_datatype,
)

logger = logging.getLogger(__name__)


class TypedIterable(Iterable[ValueType]):
    """An iterable that can infer additional type information about its elements."""

    @property
    def types(self) -> set[DataType]:
        """Collect the data types of the values in the mapping."""
        return collection_element_types(self)

    @property
    def type(self) -> DataType:
        """
        Infer the common data type of the values.
        :raise ValueError: if the iterable is empty
        :note: promotion rules: AtomicType.BOOL -> NumericType -> AtomicType.STRING
        """
        if len(self.types) == 0:
            raise ValueError("Cannot determine common type of an empty collection.")
        return common_datatype(self.types)


class Domain(set[ValueType], TypedIterable):
    """
    A set of values with additional information about the range. The collection can be freely modified, calling sort()
    caches the sorted domain.
    """

    # a sorted list of possible values
    _sorted_domain: list[ValueType] | None = None

    def invalidate(self):
        """Invalidate the cached domain information."""
        self._sorted_domain = None

    @property
    def sorted(self) -> bool:
        return self._sorted_domain is not None

    @property
    def sorted_domain(self) -> list[ValueType]:
        if self._sorted_domain is None:
            raise ValueError("the domain is not sorted")
        return self._sorted_domain

    @property
    def can_be_sorted(self) -> bool:
        try:
            sorted(self)  # type: ignore
            return True
        except TypeError:
            return False

    def sort(self):
        """Sorts the domain."""
        self._sorted_domain = None
        try:
            self._sorted_domain = sorted(self)  # type: ignore
        except TypeError:
            logger.warning("domain values cannot be sorted")

    @property
    def lower(self) -> ValueType:
        if self._sorted_domain is None:
            raise ValueError("the domain is not sorted")
        return self._sorted_domain[0]

    @property
    def upper(self) -> ValueType:
        if self._sorted_domain is None:
            raise ValueError("the domain is not sorted")
        return self._sorted_domain[-1]



"""
Utilities for (de)serializing strings.
"""


def bytes_to_string(bytestring: bytes) -> str:
    """Convert a binary string to a utf-8 string."""
    return bytestring.decode("utf-8")


def string_to_bytes(string: str) -> bytes:
    """Convert a utf-8 string to a binary string."""
    return string.encode("utf-8")



"""
Composite datatype definitions.

This module contains the core classes for defining composite datatypes (structs).
The serialization operations for composites remain in umbi.binary.composites.
"""

from dataclasses import dataclass, field

from .interval import IntervalType
from .sized_type import SizedType
from .atomic import AtomicType


@dataclass
class StructPadding:
    """Padding bits in a composite datatype."""

    padding: int

    def validate(self):
        if self.padding <= 0:
            raise ValueError(f"Padding must be positive ({self.padding})")

    @property
    def size_bits(self) -> int:
        return self.padding


@dataclass
class StructAttribute:
    """A variable field in a composite datatype."""

    name: str
    sized_type: SizedType
    is_optional: bool = (
        False  # if True, then the value is preceded with a presence flag, 1 = present, 0 = present but must be ignored
    )
    lower: int | None = None  # lower bound (for numeric types)
    upper: int | None = None  # upper bound (for numeric types)
    offset: int | None = None  # lower value offset (for numeric types)

    def validate(self):
        self.sized_type.validate()
        if isinstance(self.sized_type.type, IntervalType):
            raise ValueError(f"Struct attribute cannot be of interval type: {self.sized_type.type}")
        if self.offset is not None and self.offset != 0:
            raise NotImplementedError("non-zero offset validation not implemented")
        # TODO validate bounds for numeric types

    @property
    def size_bits(self) -> int:
        size_bits = self.sized_type.size_bits
        if self.is_optional:
            size_bits += 1  # presence flag
        return size_bits


""" Alias for either struct field type """
StructField = StructPadding | StructAttribute


@dataclass
class StructType:
    """A composite datatype consisting of attributes and paddings."""

    fields: list[StructField] = field(default_factory=list)

    def validate(self):
        for item in self.fields:
            item.validate()

    @property
    def attributes(self):
        """Iterate over attributes only, skipping padding fields."""
        return [f for f in self.fields if isinstance(f, StructAttribute)]

    @property
    def size_bits(self) -> int:
        return sum(f.size_bits for f in self.fields)

    @property
    def is_byte_aligned(self) -> bool:
        return self.size_bits % 8 == 0

    @property
    def size_bytes(self) -> int:
        num_bytes = self.size_bits // 8
        if self.size_bits % 8 != 0:
            num_bytes += 1
        return num_bytes

    @property
    def contains_strings(self) -> bool:
        return any(isinstance(f, StructAttribute) and f.sized_type.type == AtomicType.STRING for f in self.fields)

    def __str__(self) -> str:
        lines = ["struct:"]
        for f in self.fields:
            lines.append(f"  {f}")
        return "\n".join(lines)

    def __iter__(self):
        return iter(self.fields)

    def bits_to_pad(self) -> int:
        """Calculate the number of padding bits needed to align current struct to a full byte."""
        return (8 - self.size_bits % 8) % 8

    def add_padding(self, num_bits: int) -> None:
        """Add padding bits to the struct."""
        assert num_bits > 0
        self.fields.append(StructPadding(padding=num_bits))

    def pad_to_byte(self) -> None:
        """Add padding bits to align struct to a full byte."""
        if self.is_byte_aligned:
            return
        self.add_padding(self.bits_to_pad())

    def add_attribute(self, name: str, sized_type: SizedType) -> None:
        """Add an attribute field to the struct."""
        self.fields.append(StructAttribute(name=name, sized_type=sized_type))



"""
Type schema and result classes for UMBI types.
"""

from marshmallow import fields, post_load, validate
from .json_schema import JsonSchema
import umbi


class FieldType(fields.String):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            validate=validate.OneOf(
                [t.value for t in umbi.datatypes.AtomicType]
                + [t.value for t in umbi.datatypes.NumericPrimitiveType]
                + [t.value for t in umbi.datatypes.IntervalType],
                error="Must be one of: {choices}. Got: {input}",
            ),
            **kwargs,
        )

    def _deserialize(self, value, attr, data, **kwargs) -> umbi.datatypes.DataType:
        # First deserialize as string (validates against OneOf)
        string_value = super()._deserialize(value, attr, data, **kwargs)
        # Convert string to DataType enum
        for enum_class in [umbi.datatypes.AtomicType, umbi.datatypes.NumericPrimitiveType, umbi.datatypes.IntervalType]:
            try:
                return enum_class(string_value)
            except ValueError:
                pass
        raise ValueError(f"Unknown type: {string_value}")

    def _serialize(self, value, attr, obj, **kwargs) -> str:
        # Convert DataType enum to string
        if hasattr(value, "value"):
            return value.value
        return str(value)


class SizedTypeSchema(JsonSchema):
    type = FieldType(data_key="type", required=True)
    size = fields.Integer(data_key="size", required=False)

    @post_load
    def make_object(self, data: dict, **kwargs) -> umbi.datatypes.SizedType:
        obj = super().make_object(data, **kwargs)
        assert isinstance(obj.type, umbi.datatypes.DataType), f"expected DataType, got {type(obj.type)}"
        return umbi.datatypes.SizedType(type=obj.type, size_bits=obj.size)

    def dump(self, obj, *args, **kwargs):
        assert isinstance(obj, umbi.datatypes.SizedType)
        obj_dict = {
            "type": FieldType()._serialize(obj.type, "type", obj),
            "size": obj.size_bits,
        }
        return obj_dict



from importlib import import_module
from typing import TYPE_CHECKING

from .logger import set_log_level, setup_logging
from .version import __format_revision__, __format_version__, __toolname__, __version__

# Static analyzers (Pylance/mypy) evaluate this block to learn exported submodules,
# while runtime skips it to avoid eager imports and potential import cycles.
if TYPE_CHECKING:
    from . import ats as ats
    from . import binary as binary
    from . import datatypes as datatypes
    from . import index as index
    from . import io as io
    from . import version as version


_SUBMODULES = {"ats", "binary", "datatypes", "index", "io", "version"}


def __getattr__(name: str):
    if name in _SUBMODULES:
        module = import_module(f".{name}", __name__)
        globals()[name] = module
        return module
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    return sorted(list(globals().keys()) + list(_SUBMODULES))


__all__ = [
    "set_log_level",
    "setup_logging",
    "__toolname__",
    "__version__",
    "__format_version__",
    "__format_revision__",
    "datatypes",
    "binary",
    "io",
    "index",
    "ats",
    "version",
]



import pathlib
import time

import umbi

from .umb import ExplicitUmb, read_umb, write_umb


def umbi_file_data() -> umbi.index.FileData:
    """Generate UMB file data for use in umb files created by umbi."""
    return umbi.index.FileData(
        tool=umbi.version.__toolname__,
        tool_version=umbi.version.__version__,
        creation_date=int(time.time()),
        # parameters=parameters,
    )


def explicit_umb_to_explicit_ats(umb: ExplicitUmb) -> umbi.ats.ExplicitAts:
    umb.validate()
    explicit_ats = umbi.ats.ExplicitAts()

    ## index
    # skip format_version, format_revision and file_data
    md = umb.index.model_data
    if md is not None:
        explicit_ats.model_info = umbi.ats.ModelInfo(
            name=md.name,
            version=md.version,
            authors=md.authors,
            description=md.description,
            comment=md.comment,
            doi=md.doi,
            url=md.url,
        )
    # load index.transition_system
    ts = umb.index.transition_system
    explicit_ats.time = umbi.ats.TimeType(ts.time)
    explicit_ats.num_players = ts.num_players
    explicit_ats.num_states = ts.num_states
    explicit_ats.num_initial_states = ts.num_initial_states
    explicit_ats.num_choices = ts.num_choices
    explicit_ats.num_choice_actions = ts.num_choice_actions
    explicit_ats.num_branches = ts.num_branches
    explicit_ats.num_branch_actions = ts.num_branch_actions
    explicit_ats.player_to_name = ts.player_names

    ## values
    explicit_ats.state_is_initial = umb.state_is_initial
    explicit_ats.state_to_choice = umb.state_to_choices
    explicit_ats.state_to_player = umb.state_to_player

    explicit_ats.state_is_markovian = umb.state_is_markovian
    explicit_ats.state_to_exit_rate = umb.state_to_exit_rate

    explicit_ats.choice_to_branches = umb.choice_to_branches
    explicit_ats.branch_to_target = umb.branch_to_target
    explicit_ats.branch_to_probability = umb.branch_to_probability

    explicit_ats.choice_to_choice_action = umb.choice_to_choice_action
    explicit_ats.choice_action_to_name = umb.choice_action_to_string
    explicit_ats.branch_to_branch_action = umb.branch_to_branch_action
    explicit_ats.branch_action_to_name = umb.branch_action_to_string

    # load annotations
    if umb.index.annotations is not None:
        assert umb.annotations is not None
        for category, name_to_annotation in umb.index.annotations.items():
            constructor = {
                "rewards": umbi.ats.RewardAnnotation,
                "aps": umbi.ats.AtomicPropositionAnnotation,
            }[category] or umbi.ats.Annotation
            for name, umb_annotation in name_to_annotation.items():
                ats_annotation = constructor(
                    name=name,
                    alias=umb_annotation.alias,
                    description=umb_annotation.description,
                )
                for applies_to in umb_annotation.applies_to or []:
                    values = umb.annotations[category][name][applies_to]
                    entity_class = umbi.ats.EntityClass(applies_to)
                    ats_annotation.set_values_for(entity_class, values)
                explicit_ats.annotations[category] = dict[str, umbi.ats.Annotation]()
                explicit_ats.annotations[category][name] = ats_annotation

    # load valuations
    if umb.index.valuations is not None:
        assert umb.valuations is not None
        explicit_ats.variable_valuations = umbi.ats.EntityClassValuations()
        for applies_to, valuation_description in umb.index.valuations.items():
            # ignore unique and num_strings
            # assume a single valuation class
            if len(valuation_description.classes) > 1:
                raise NotImplementedError("multiple valuation classes not supported yet")
            struct_type = valuation_description.classes[0]

            entity_to_valuation = umb.valuations[applies_to]
            entity_class = umbi.ats.EntityClass(applies_to)
            entity_valuations = umbi.ats.EntityValuations()
            for attribute in struct_type.attributes:
                if attribute.lower is not None or attribute.upper is not None:
                    raise NotImplementedError("bounds on valuation variables not supported yet")
                if attribute.offset is not None and attribute.offset != 0:
                    raise NotImplementedError("offsets on valuation variables not supported yet")
                entity_valuations.add_variable(variable_name=attribute.name)
            for entity, valuation in enumerate(entity_to_valuation):
                valuation = {entity_valuations.get_variable(var_name): value for var_name, value in valuation.items()}
                entity_valuations.set_entity_valuation(entity, valuation)
            explicit_ats.variable_valuations.set_valuations_for(entity_class, entity_valuations)

    # load observations
    if ts.observations_apply_to is not None:
        assert ts.num_observations > 0
        assert umb.entity_to_observation is not None
        explicit_ats.observation_annotation = umbi.ats.ObservationAnnotation(num_observations=ts.num_observations)
        entity_class = umbi.ats.EntityClass(ts.observations_apply_to)
        explicit_ats.observation_annotation.set_values_for(entity_class, umb.entity_to_observation)

    explicit_ats.validate()
    return explicit_ats


def explicit_ats_to_explicit_umb(ats: umbi.ats.ExplicitAts) -> ExplicitUmb:
    ats.validate()
    umb = ExplicitUmb()

    ## index

    umb.index = umbi.index.UmbIndex(
        # insert our format_version, format_revision, file data
        format_version=umbi.version.__format_version__,
        format_revision=umbi.version.__format_revision__,
        file_data=umbi_file_data(),
        # add model data
        model_data=(
            umbi.index.ModelData(
                name=ats.model_info.name,
                version=ats.model_info.version,
                authors=ats.model_info.authors,
                description=ats.model_info.description,
                comment=ats.model_info.comment,
                doi=ats.model_info.doi,
                url=ats.model_info.url,
            )
            if ats.model_info is not None
            else None
        ),
        transition_system=umbi.index.TransitionSystem(
            time=ats.time.value,
            num_players=ats.num_players,
            num_states=ats.num_states,
            num_initial_states=ats.num_initial_states,
            num_choices=ats.num_choices,
            num_choice_actions=ats.num_choice_actions,
            num_branches=ats.num_branches,
            num_branch_actions=ats.num_branch_actions,
            num_observations=ats.num_observations,
            observations_apply_to=None,  # for now
            branch_probability_type=None,  # for now
            exit_rate_type=None,  # for now
            observation_probability_type=None,  # for now
            player_names=ats.player_to_name,
        ),
        # create annotations
        annotations=None,  # for now
        valuations=None,  # for now
    )

    ## values
    # warning: this does not copy the values, but directly uses the lists from ats.
    # this is fine as long as we don't modify them
    umb.state_is_initial = ats.state_is_initial
    umb.state_to_choices = ats.state_to_choice
    umb.state_to_player = ats.state_to_player

    umb.state_is_markovian = ats.state_is_markovian
    if ats.state_to_exit_rate is not None:
        # promote all to common type
        target_type = umbi.datatypes.common_collection_element_type(ats.state_to_exit_rate)
        assert isinstance(target_type, umbi.datatypes.NumericType), "exit rates must be numeric"
        # do we need to promote ints?
        # if target_type == umbi.datatypes.NumericPrimitiveType.INT:
        #     # implicit promotion of int to double
        #     target_type = umbi.datatypes.NumericPrimitiveType.DOUBLE
        umb.index.transition_system.exit_rate_type = umbi.datatypes.SizedType(type=target_type)
        umb.state_to_exit_rate = umbi.datatypes.promote_vector_to(ats.state_to_exit_rate, target_type)  # type: ignore

    umb.choice_to_branches = ats.choice_to_branches
    umb.branch_to_target = ats.branch_to_target

    if ats.branch_to_probability is not None:
        # promote
        target_type = umbi.datatypes.common_collection_element_type(ats.branch_to_probability)
        assert isinstance(target_type, umbi.datatypes.NumericType), "branch probabilities must be numeric"
        umb.index.transition_system.branch_probability_type = umbi.datatypes.SizedType(type=target_type)
        umb.branch_to_probability = umbi.datatypes.promote_vector_to(ats.branch_to_probability, target_type)  # type: ignore

    umb.choice_to_choice_action = ats.choice_to_choice_action
    umb.choice_action_to_string = ats.choice_action_to_name
    umb.branch_to_branch_action = ats.branch_to_branch_action
    umb.branch_action_to_string = ats.branch_action_to_name

    # add annotations
    umb.index.annotations = {}
    umb.annotations = {}
    for category, name_to_annotation in ats.annotations.items():
        if len(name_to_annotation) == 0:
            continue
        umb.index.annotations[category] = dict[str, umbi.index.Annotation]()
        umb.annotations[category] = dict[str, dict[str, list]]()
        for name, ats_annotation in name_to_annotation.items():
            # determine common type of values for this annotation
            types = {
                umbi.datatypes.common_collection_element_type(values)
                for values in ats_annotation._entity_class_to_values.values()
            }
            target_type = umbi.datatypes.common_datatype(types)
            umb_annotation = umbi.index.Annotation(
                alias=ats_annotation.alias,
                description=ats_annotation.description,
                applies_to=[entity_class.value for entity_class in ats_annotation.entity_classes],  # type: ignore
                type=umbi.datatypes.SizedType(type=target_type),
                lower=None,  # TODO add later
                upper=None,  # TODO add later
            )
            umb.index.annotations[category][name] = umb_annotation
            umb.annotations[category][name] = dict[str, list]()
            for entity_class in ats_annotation.entity_classes:
                values = ats_annotation._entity_class_to_values[entity_class]
                values = umbi.datatypes.promote_vector_to(values, target_type)  # type: ignore
                umb.annotations[category][name][entity_class.value] = values

    # add valuations
    if ats.variable_valuations is not None:
        umb.index.valuations = {}
        umb.valuations = {}
        for entity_class, entity_valuations in ats.variable_valuations.items():
            applies_to = entity_class.value
            valuation_class = umbi.datatypes.StructType()
            for var in entity_valuations.variables:
                valuation_class.add_attribute(
                    name=var.name,
                    sized_type=umbi.datatypes.SizedType(type=var.type),  # TODO deduce non-default size from the values
                )
            valuation_class.pad_to_byte()
            valuation_description = umbi.index.ValuationDescription(
                unique=False,  # TODO add later
                num_strings=None,  # TODO add later
                classes=[valuation_class],
            )
            umb.index.valuations[applies_to] = valuation_description
            umb.valuations[applies_to] = []
            for entity in range(entity_valuations.num_entities):
                valuations = entity_valuations.get_entity_valuation(entity)
                valuations = {var.name: value for var, value in valuations.items()}
                umb.valuations[applies_to].append(valuations)

    # add observations
    if ats.observation_annotation is not None:
        umb.index.transition_system.observations_apply_to = ats.observation_annotation.entity_class.value  # type: ignore
        umb.index.transition_system.num_observations = ats.observation_annotation.num_observations
        umb.entity_to_observation = ats.observation_annotation.values

    umb.validate()
    return umb


def read_ats(umbpath: str | pathlib.Path) -> umbi.ats.ExplicitAts:
    """Read ATS from a umbfile."""
    umb = read_umb(umbpath)
    ats = explicit_umb_to_explicit_ats(umb)
    return ats


def write_ats(ats: umbi.ats.ExplicitAts, umbpath: str | pathlib.Path) -> None:
    """Write ATS to a umbfile."""
    umb = explicit_ats_to_explicit_umb(ats)
    write_umb(umb, umbpath)



"""
File data schemas and classes.
"""

from dataclasses import dataclass

from marshmallow import fields

import umbi

from .json_schema import (
    JsonSchema,
    JsonSchemaResult,
    FieldUint,
)


class FileDataSchema(JsonSchema):
    """File data schema."""

    tool = fields.String(data_key="tool", required=False)
    tool_version = fields.String(data_key="tool-version", required=False)
    creation_date = FieldUint(data_key="creation-date", required=False)
    parameters = fields.Raw(data_key="parameters", required=False)

    @classmethod
    def schema_class(cls) -> type:
        return FileData


@dataclass
class FileData(JsonSchemaResult):
    """File data class."""

    tool: str | None = None
    tool_version: str | None = None
    creation_date: int | None = None
    parameters: umbi.datatypes.JsonLike | None = None

    @classmethod
    def class_schema(cls) -> type:
        return FileDataSchema



import logging
import click
from umbi import setup_logging, set_log_level, __toolname__, __version__, io

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    default="INFO",
    show_default=True,
    required=False,
    help="logging level",
)
@click.option("--import-umb", type=click.Path(), required=False, help=".umb filepath to import")
@click.option("--export-umb", type=click.Path(), required=False, help=".umb filepath to export")
def main(log_level, import_umb, export_umb):
    setup_logging()
    set_log_level(getattr(logging, log_level))
    logger.info(f"this is {__toolname__} v.{__version__}")

    ats = None
    if import_umb is not None:
        ats = io.read_umb(import_umb)
    if export_umb is not None:
        if ats is None:
            raise ValueError("--export-umb specified, but nothing to export")
        io.write_umb(ats, export_umb)


if __name__ == "__main__":
    main()



from dataclasses import dataclass, field
from umbi.datatypes import (
    AtomicType,
    DataType,
    NumericType,
    NumericPrimitiveType,
    common_datatype,
)

from .entity_class import EntityClass
from .custom_lists import TypedIterable
import logging

logger = logging.getLogger(__name__)


class EntityToValue(list, TypedIterable):
    """A list of values associated with entities of a certain class."""

    pass


@dataclass
class Annotation:
    """General annotation."""

    name: str
    alias: str | None = None
    description: str | None = None
    _entity_class_to_values: dict[EntityClass, EntityToValue] = field(default_factory=dict[EntityClass, EntityToValue])

    @classmethod
    def entity_class_enabled(cls, entity_class: EntityClass) -> bool:
        """Check if the given entity class is enabled for this annotation type. Can be overridden by subclasses."""
        return True

    @property
    def mappings(self) -> dict[EntityClass, EntityToValue]:
        """Alias for entity_class_to_values."""
        return self._entity_class_to_values

    @property
    def has_values(self) -> bool:
        """Check if the annotation has any values set."""
        return len(self._entity_class_to_values) > 0

    @property
    def entity_classes(self) -> set[EntityClass]:
        """Get the set of entity classes for which this annotation has values."""
        return set(self._entity_class_to_values.keys())

    def has_values_for(self, entity_class: EntityClass) -> bool:
        """Check if the annotation has values for the given entity class."""
        return self._entity_class_to_values.get(entity_class) is not None

    def get_values_for(self, entity_class: EntityClass) -> EntityToValue:
        """
        Get the values for the given entity class.
        :raises KeyError: if no values are set for the given entity class
        """
        if not self.has_values_for(entity_class):
            raise KeyError(f"Annotation has no values for entity class {entity_class}")
        return self._entity_class_to_values[entity_class]

    def set_values_for(self, entity_class: EntityClass, values: EntityToValue | list) -> None:
        """Set the values for the given entity class."""
        if not self.entity_class_enabled(entity_class):
            raise ValueError(f"Entity class {entity_class} is not enabled for this annotation type")
        if isinstance(values, list):
            values = EntityToValue(values)
        self._entity_class_to_values[entity_class] = values

    def unset_values_for(self, entity_class: EntityClass) -> None:
        """Unset the values for the given entity class."""
        if not self.has_values_for(entity_class):
            raise KeyError(f"Annotation has no values for entity class {entity_class}")
        del self._entity_class_to_values[entity_class]

    ### Convenience properties and methods for each entity class ###
    @property
    def has_state_values(self) -> bool:
        return self.has_values_for(EntityClass.STATES)

    @property
    def has_choice_values(self) -> bool:
        return self.has_values_for(EntityClass.CHOICES)

    @property
    def has_branch_values(self) -> bool:
        return self.has_values_for(EntityClass.BRANCHES)

    @property
    def has_observation_values(self) -> bool:
        return self.has_values_for(EntityClass.OBSERVATIONS)

    @property
    def has_player_values(self) -> bool:
        return self.has_values_for(EntityClass.PLAYERS)

    @property
    def state_values(self) -> EntityToValue:
        return self.get_values_for(EntityClass.STATES)

    @property
    def choice_values(self) -> EntityToValue:
        return self.get_values_for(EntityClass.CHOICES)

    @property
    def branch_values(self) -> EntityToValue:
        return self.get_values_for(EntityClass.BRANCHES)

    @property
    def observation_values(self) -> EntityToValue:
        return self.get_values_for(EntityClass.OBSERVATIONS)

    @property
    def player_values(self) -> EntityToValue:
        return self.get_values_for(EntityClass.PLAYERS)

    def set_state_values(self, values: EntityToValue | list):
        self.set_values_for(EntityClass.STATES, values)

    def set_choice_values(self, values: EntityToValue | list):
        self.set_values_for(EntityClass.CHOICES, values)

    def set_branch_values(self, values: EntityToValue | list):
        self.set_values_for(EntityClass.BRANCHES, values)

    def set_observation_values(self, values: EntityToValue | list):
        self.set_values_for(EntityClass.OBSERVATIONS, values)

    def set_player_values(self, values: EntityToValue | list):
        self.set_values_for(EntityClass.PLAYERS, values)

    def unset_state_values(self):
        self.unset_values_for(EntityClass.STATES)

    def unset_choice_values(self):
        self.unset_values_for(EntityClass.CHOICES)

    def unset_branch_values(self):
        self.unset_values_for(EntityClass.BRANCHES)

    def unset_observation_values(self):
        self.unset_values_for(EntityClass.OBSERVATIONS)

    def unset_player_values(self):
        self.unset_values_for(EntityClass.PLAYERS)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Annotation):
            return False
        return (
            self.name == other.name
            and self.alias == other.alias
            and self.description == other.description
            and self._entity_class_to_values == other._entity_class_to_values
        )

    @property
    def common_type(self) -> DataType:
        """Infer the data type of the annotation from its values."""
        # TODO this is obsolete, get rid of type handling in this package
        types: set[DataType] = set()
        for v in self._entity_class_to_values.values():
            types.add(v.type)
        if len(types) == 0:
            raise ValueError("Annotation has no values to infer type from")
        return common_datatype(types)

    def validate(self):
        """Validate the annotation data."""
        if not isinstance(self.name, str):
            raise ValueError("name must be a string")
        if not isinstance(self.alias, (str, type(None))):
            raise ValueError("alias must be a string or None")
        if not isinstance(self.description, (str, type(None))):
            raise ValueError("description must be a string or None")
        return


class RewardAnnotation(Annotation):
    """Reward annotation is a numeric annotation that can be associated with states, choices, or branches."""

    @classmethod
    def entity_class_enabled(cls, entity_class: EntityClass) -> bool:
        return entity_class in {
            EntityClass.STATES,
            EntityClass.CHOICES,
            EntityClass.BRANCHES,
        }

    def validate(self):
        super().validate()
        # Additional validation for reward annotations can be added here
        type = self.common_type
        if not isinstance(type, NumericType):
            raise ValueError(f"Reward annotation type must be NumericType, got {type}")


class AtomicPropositionAnnotation(Annotation):
    """Atomic proposition annotation is a boolean annotation that can only be associated with states."""

    @classmethod
    def entity_class_enabled(cls, entity_class: EntityClass) -> bool:
        return entity_class == EntityClass.STATES

    def validate(self):
        super().validate()
        # Additional validation for AP annotations can be added here
        type = self.common_type
        if type != AtomicType.BOOL:
            raise ValueError(f"AP annotation type must be AtomicType.BOOL, got {type}")


@dataclass
class ObservationAnnotation(Annotation):
    """Observation annotation is an integer annotation that can only be associated with either states or branches."""

    def __init__(self, num_observations: int):
        super().__init__(name="observation")
        self.num_observations = num_observations

    @classmethod
    def entity_class_enabled(cls, entity_class: EntityClass) -> bool:
        return entity_class in {EntityClass.STATES, EntityClass.BRANCHES}

    @property
    def entity_class(self) -> EntityClass:
        """Get the entity class for which this observation annotation has values."""
        if not self.has_values:
            raise ValueError("Observation annotation does not have any values set")
        return next(iter(self.entity_classes))

    @property
    def values(self) -> EntityToValue:
        """Get the values for the entity class of this observation annotation."""
        return self.get_values_for(self.entity_class)

    def set_values_for(self, entity_class: EntityClass, values: EntityToValue | list):
        if self.has_values and entity_class != self.entity_class:
            raise ValueError("Observation annotation can only have values for one entity class")
        return super().set_values_for(entity_class, values)

    def validate(self) -> None:
        super().validate()
        if not (isinstance(self.num_observations, int) and 0 < self.num_observations):
            raise ValueError(f"num_observations must be a positive integer, got {self.num_observations}")
        if self.common_type != NumericPrimitiveType.INT:
            raise ValueError(f"Observation annotation type must be INT, got {self.common_type}")
        for item, obs in enumerate(self.values):
            if not 0 <= obs < self.num_observations:  # type: ignore
                raise ValueError(f"observation mapping[{item}] = {obs} is out of range [0, {self.num_observations})")

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self.num_observations == other.num_observations



"""
Common datatype for all primitive and interval types.
"""

from .atomic import Atomic, AtomicType, atomic_type_of
from .numeric import NumericType, Numeric, numeric_type_of, common_numeric_type, promote_numeric_to

""" Alias for all common types. """
DataType = AtomicType | NumericType

""" Alias for all common objects. """
ValueType = Atomic | Numeric


def datatype_of(value: ValueType) -> DataType:
    """Determine the common or interval type of a given value."""
    if isinstance(value, Atomic):
        return atomic_type_of(value)
    else:  # isinstance(value, Numeric):
        return numeric_type_of(value)


def common_datatype(types: set[DataType]) -> DataType:
    """
    Determine the common type from a set of types. Used for type promotion.
    Rules: bool -> Numeric -> string
    """
    assert len(types) > 0, "cannot determine common type of empty set"
    assert all(isinstance(t, DataType) for t in types), f"non-common types found in set: {types}"
    if AtomicType.STRING in types:
        return AtomicType.STRING
    if AtomicType.BOOL in types:
        if len(types) == 1:
            return AtomicType.BOOL
        types.remove(AtomicType.BOOL)
    return common_numeric_type(types)  # type: ignore


def promote_value_to(value: ValueType, target_type: DataType) -> ValueType:
    """Promote a value to the target type."""
    if target_type == AtomicType.STRING:
        return str(value)
    elif target_type == AtomicType.BOOL:
        assert isinstance(value, bool), f"cannot promote {value} to bool"
        return value
    else:  # isinstance(target_type, NumericType):
        if isinstance(value, bool):
            value = int(value)
        assert isinstance(value, Numeric), f"cannot promote {value} to numeric type {target_type}"
        return promote_numeric_to(value, target_type)
